/*===========================================================================*/
/*= UTILITY ROUTINES (Non-Exportable Functions) =============================*/
/*===========================================================================*/
#ifdef _WIN32
  #ifdef _CVI_
    #include <userint.h>
  #else
    #include <windows.h>
  #endif
#endif
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "rssmf.h"
#include "rssmf_utility.h"

#define IEEE_ERROR_BITS         0x3C    /* AND mask for ESR checking */
#define BUFFER_SIZE             2048UL    /* File I/O buffer size */
#define RS_IO_BUFFER_SIZE               1048576UL   /* I/O buffer size */

/*===========================================================================*/
/* Function: Boolean Value Out Of Range - ViBoolean                          */
/* Purpose:  This function checks a Boolean to see if it is equal to VI_TRUE */
/*           or VI_FALSE. If the value is out of range, the return value is  */
/*           VI_TRUE, otherwise the return value is VI_FALSE.                */
/*===========================================================================*/
ViBoolean rssmf_invalidViBooleanRange (ViSession instrSession, ViBoolean val)
{
    rssmf_instrRange    instrPtr = NULL;

    viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr);

    /* Error Checking Enabled? */
    if (instrPtr != NULL)
    {
        if (instrPtr -> rangeChecking != VI_FALSE)
        {
            return(ViBoolean)((val != VI_FALSE && val != VI_TRUE) ? VI_TRUE : VI_FALSE);
        }
    }
    return VI_FALSE;
}

/*===========================================================================*/
/* Function: Long Signed Integer Value Out Of Range - ViInt32                */
/* Purpose:  This function checks a long signed integer value to see if it   */
/*           lies between a minimum and maximum value.  If the value is out  */
/*           of range, the return value is VI_TRUE, otherwise the return     */
/*           value is VI_FALSE.                                              */
/*===========================================================================*/
ViBoolean rssmf_invalidViInt32Range (ViSession instrSession, ViInt32 val, ViInt32 min, ViInt32 max)
{
    rssmf_instrRange    instrPtr = NULL;

    viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr);

    /* Error Checking Enabled? */
    if (instrPtr != NULL)
    {
        if (instrPtr -> rangeChecking != VI_FALSE)
        {
            return (ViBoolean)((val < min || val > max) ? VI_TRUE : VI_FALSE);
        }
    }
    return VI_FALSE;
}

/*===========================================================================*/
/* Function: Long Unsigned Integer Value Out Of Range - ViUInt32             */
/* Purpose:  This function checks a long signed integer value to see if it   */
/*           lies between a minimum and maximum value.  If the value is out  */
/*           of range, the return value is VI_TRUE, otherwise the return     */
/*           value is VI_FALSE.                                              */
/*===========================================================================*/
ViBoolean rssmf_invalidViUInt32Range (ViSession instrSession, ViUInt32 val, ViUInt32 min, ViUInt32 max)
{
    rssmf_instrRange    instrPtr = NULL;

    viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr);

    /* Error Checking Enabled? */
    if (instrPtr != NULL)
    {
        if (instrPtr -> rangeChecking != VI_FALSE)
        {
            (ViBoolean)((val < min || val > max) ? VI_TRUE : VI_FALSE);
        }
    }
    return VI_FALSE;
}

/*===========================================================================*/
/* Function: Real (Double) Value Out Of Range - ViReal64                     */
/* Purpose:  This function checks a real (double) value to see if it lies    */
/*           between a minimum and maximum value.  If the value is out of    */
/*           range, the return value is VI_TRUE, otherwise the return value  */
/*           is VI_FALSE.                                                    */
/*===========================================================================*/
ViBoolean rssmf_invalidViReal64Range (ViSession instrSession, ViReal64 val, ViReal64 min, ViReal64 max)
{
    rssmf_instrRange    instrPtr = NULL;

    viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr);

    /* Error Checking Enabled? */
    if (instrPtr != NULL)
    {
        if (instrPtr -> rangeChecking != VI_FALSE)
        {
            return (ViBoolean)((val < min || val > max) ? VI_TRUE : VI_FALSE);
        }
    }
    return VI_FALSE;
}
/*=========================================================================*/
/* Function: Check Status                                                  */
/* Purpose:  This function reads Event Status Register to check for        */
/*           possible errors.                                              */
/*=========================================================================*/
ViStatus rssmf_checkStatus (ViSession instrSession)
{
    ViStatus            rssmf_status = VI_SUCCESS;
    rssmf_instrRange    instrPtr = NULL;

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));

    /* Error Checking Enabled? */
    if (instrPtr != NULL)
    {
        if (instrPtr -> errorChecking != VI_FALSE)
        {
            ViInt32 esrByte = 0;

            /*--------------- Read the Standard Event Register ------------------*/
            CHECKERR (viQueryf (instrSession, "*ESR?\n", "%ld", &esrByte));

            /*---------------- Check if any error or message bit was asserted ---*/
            rssmf_status = ((esrByte & IEEE_ERROR_BITS) != 0) ? RSSMF_ERROR_INSTRUMENT_ERROR : VI_SUCCESS;
        }
    }

    return rssmf_status;
}

/*===========================================================================*/
/* Function: Wait for OPC                                                    */
/* Purpose:  This function sends a command to the instrument and waits until */
/*           the instrument response is available (queries MAV in STB).      */
/*===========================================================================*/
ViStatus rssmf_waitOPC (ViSession instrSession, ViPString command,
                ViPString instrResp, ViInt32 timeout)
{
    ViStatus    rssmf_status    = VI_SUCCESS;
    ViUInt16    response        = 0;
    ViInt32     counter         = 0;
    ViInt32     oldTimeout      = 0;
    ViUInt16    intf       = VI_INTF_GPIB;

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_INTF_TYPE, &intf));

    /* Interface specific operations */
    switch (intf)
        {
        case VI_INTF_ASRL: /* Serial intf */
            CHECKERR (viGetAttribute (instrSession, VI_ATTR_TMO_VALUE, &oldTimeout));
            CHECKERR (viSetAttribute (instrSession, VI_ATTR_TMO_VALUE, (ViAttrState) timeout));
            if ((rssmf_status = viQueryf (instrSession, "%s\n", "%[^\r\n]", command, instrResp)) < 0)
            {
                (void) viSetAttribute (instrSession, VI_ATTR_TMO_VALUE, (ViAttrState) oldTimeout);
                return rssmf_status;
            }
            CHECKERR (viSetAttribute (instrSession, VI_ATTR_TMO_VALUE, (ViAttrState) oldTimeout));
        break;
        default: /* Default case */
            CHECKERR (viPrintf (instrSession, "%s\n", command));

            while ((response & 0x10) == 0) /* (queries MAV in STB) */
                {

                CHECKERR (viReadSTB (instrSession, &response));

                (void) rssmf_delay (instrSession, 0.005);
                counter+= 5;
                if (counter >= timeout)
                    return VI_ERROR_TMO;
                }

            CHECKERR (viScanf (instrSession, "%[^\r\n]", instrResp));
        break;
        }

    return rssmf_status;
}

/*===========================================================================*/
/* Function: Send Command                                                    */
/* Purpose:  This function sends command to the instrument with error check. */
/*===========================================================================*/
ViStatus rssmf_sendCmd (ViSession instrSession, ViString cmd)
{
    ViStatus    rssmf_status    = VI_SUCCESS;
    size_t      length          = 0;
    ViChar      *pbuffer;

    length = strlen (cmd);
    pbuffer = (ViChar*) malloc(length + 1);
    if (pbuffer == NULL)
        return VI_ERROR_ALLOC;

    memcpy (pbuffer, cmd, length);
    pbuffer[length] = '\n';

    length++;
    GOCHECKERR (viWrite (instrSession, (ViPBuf) cmd, (ViUInt32) length, VI_NULL));
    GOCHECKERR (rssmf_checkStatus (instrSession));

ERROR_LABEL:

    free(pbuffer);

    return rssmf_status;
}

/*===========================================================================*/
/* Function: Set ViInt32                                                     */
/* Purpose:  This function sets command with integer parameter.              */
/*===========================================================================*/
ViStatus rssmf_setViInt32 (ViSession instrSession,
                           ViString cmd,
                           ViInt32 value,
                           ViInt32 min,
                           ViInt32 max,
                           ViString parameterName)
{
    ViChar      *wrt_buf;
    ViStatus    rssmf_status = VI_SUCCESS;

    if ((max - min) > 0)
        if (rssmf_invalidViInt32Range (instrSession, value, min, max) == VI_TRUE)
		{
		    rssmf_instrRange   instrPtr;
		    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
			strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
	        return VI_ERROR_INSTR_PARAMETER;
		}

    
    wrt_buf = (ViChar*) malloc (strlen (cmd) + 20);
    
    sprintf (wrt_buf, "%s %ld", cmd, value);

    rssmf_status = rssmf_sendCmd (instrSession, wrt_buf);
    
    free (wrt_buf);
    return rssmf_status;
}

/*===========================================================================*/
/* Function: Get ViInt32                                                     */
/* Purpose:  This function gets command with integer parameter.              */
/*===========================================================================*/
ViStatus rssmf_getViInt32 (ViSession instrSession,
                           ViString cmd,
                           ViInt32 *value,
                           ViString parameterName)
{
    ViStatus    rssmf_status = VI_SUCCESS;

    if (value == VI_NULL)
	{
	    rssmf_instrRange   instrPtr;
	    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
        return VI_ERROR_INSTR_PARAMETER;
	}

    CHECKERR( viPrintf(instrSession, "%s\n", cmd));
    CHECKERR( viScanf(instrSession, "%ld", value));
    CHECKERR( rssmf_checkStatus (instrSession));

    return rssmf_status;
}

/*===========================================================================*/
/* Function: Set ViBoolean                                                   */
/* Purpose:  This function sets command with integer parameter.              */
/*===========================================================================*/
ViStatus rssmf_setViBoolean (ViSession instrSession,
                           ViString cmd,
                           ViBoolean value,
                           ViString parameterName)
{
    ViStatus rssmf_status = VI_SUCCESS;
	if (rssmf_invalidViBooleanRange (instrSession, value) == VI_TRUE)
	{
	    rssmf_instrRange   instrPtr;
	    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
        return VI_ERROR_INSTR_PARAMETER;
	}

    return rssmf_setViString (instrSession,
                              cmd,
                              (value == VI_TRUE) ? "ON" : "OFF",
                              0);
}

/*===========================================================================*/
/* Function: Get ViBoolean                                                   */
/* Purpose:  This function gets command with integer parameter.              */
/*===========================================================================*/
ViStatus rssmf_getViBoolean (ViSession instrSession,
                           ViString cmd,
                           ViBoolean *value,
                           ViString parameterName)
{
    ViStatus    rssmf_status = VI_SUCCESS;
    ViChar      buffer[4];

    if (value == NULL)
	{
	    rssmf_instrRange   instrPtr;
	    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
        return VI_ERROR_INSTR_PARAMETER;
	}

    CHECKERR (rssmf_getViString (instrSession,
                                cmd,
                                buffer, 4,
                                0));

    *value = (ViBoolean) ((strncmp (buffer, "1", 1) == 0) || (strncmp (buffer,"ON",2) == 0));

    return rssmf_status;
}

/*===========================================================================*/
/* Function: Set ViBoolean With Path                                         */
/* Purpose:  This function sets command with integer parameter.              */
/*===========================================================================*/
ViStatus rssmf_setViBooleanPath (ViSession instrSession,
                           ViString cmd,
                           ViInt32 path,
                           ViBoolean value,
                           ViString parameterName)
{
    ViStatus    rssmf_status = VI_SUCCESS;
    ViChar  wrt_buf[BUFFER_SIZE];
    ViChar* p2buf;

    CHECKPATH;
    if (rssmf_invalidViBooleanRange (instrSession, value) != 0)
	{
	    rssmf_instrRange   instrPtr;
	    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
        return VI_ERROR_INSTR_PARAMETER;
	}

    p2buf = wrt_buf + sprintf(wrt_buf,cmd, path);
    sprintf (p2buf, " %s", (value != 0) ? "ON" : "OFF");

    return rssmf_sendCmd (instrSession, wrt_buf);
}

/*===========================================================================*/
/* Function: Get ViBoolean With Path                                         */
/* Purpose:  This function gets command with integer parameter.              */
/*===========================================================================*/
ViStatus rssmf_getViBooleanPath (ViSession instrSession,
                           ViString cmd,
                           ViInt32 path,
                           ViBoolean *value,
                           ViString parameterName)
{
    ViChar  wrt_buf[BUFFER_SIZE];

    CHECKPATH;

    sprintf(wrt_buf,cmd, path);

    return rssmf_getViBoolean(instrSession,
                                   wrt_buf,
                                   value,
                                   parameterName);
}

/*===========================================================================*/
/* Function: Set ViReal64                                                    */
/* Purpose:  This function sets command with double parameter.               */
/*===========================================================================*/
ViStatus rssmf_setViReal64 (ViSession instrSession,
                            ViString cmd,
                            ViReal64 value,
                            ViReal64 min,
                            ViReal64 max,
                            ViString parameterName)
{
    ViChar      buffer[BUFFER_SIZE];
    ViStatus    rssmf_status = VI_SUCCESS;

    /* do range checking only if there's difference between max and min */
    if ((max - min) > 1.0e-14)
        if (rssmf_invalidViReal64Range (instrSession, value, min, max) == VI_TRUE)
		{
		    rssmf_instrRange   instrPtr;
		    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
			strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
	        return VI_ERROR_INSTR_PARAMETER;
		}

    sprintf(buffer,"%s %.12lG", cmd, value);

    return rssmf_sendCmd(instrSession, buffer);
}

/*===========================================================================*/
/* Function: Get ViReal64                                                    */
/* Purpose:  This function gets command with double parameter.               */
/*===========================================================================*/
ViStatus rssmf_getViReal64 (ViSession instrSession,
                            ViString cmd,
                            ViReal64 *value,
                            ViString parameterName)
{
    ViStatus    rssmf_status = VI_SUCCESS;

    if ( value == VI_NULL )
	{
	    rssmf_instrRange   instrPtr;
	    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
        return VI_ERROR_INSTR_PARAMETER;
	}

    CHECKERR(viPrintf(instrSession, "%s\n",cmd));
    CHECKERR(viScanf (instrSession, "%le",value));
    CHECKERR(rssmf_checkStatus (instrSession));

    return rssmf_status;
}

/*===========================================================================*/
/* Function: Set ViReal64 With Channel                                       */
/* Purpose:  This function sets command with double parameter.               */
/*===========================================================================*/
ViStatus rssmf_setViReal64Path (ViSession instrSession,
                            ViString cmd,
                            ViInt32  path,
                            ViReal64 value,
                            ViReal64 min,
                            ViReal64 max,
                            ViString parameterName)
{
    ViChar      wrt_buf[BUFFER_SIZE];

    CHECKPATH;

    sprintf(wrt_buf,cmd, path);

    return rssmf_setViReal64(instrSession,
                             wrt_buf,
                             value,
                             min,
                             max,
                             parameterName);
}

/*===========================================================================*/
/* Function: Get ViReal64 With Channel                                       */
/* Purpose:  This function gets command with double parameter.               */
/*===========================================================================*/
ViStatus rssmf_getViReal64Path (ViSession instrSession,
                            ViString cmd,
                            ViInt32 path,
                            ViReal64 *value,
                            ViString parameterName)
{
    ViChar      wrt_buf[BUFFER_SIZE];

    CHECKPATH;

    sprintf(wrt_buf, cmd, path);

    return rssmf_getViReal64(instrSession,
                             wrt_buf,
                             value,
                             parameterName);
}

/*===========================================================================*/
/* Function: Set ViReal64 with Units                                         */
/* Purpose:  This function sets command with double parameter.               */
/*===========================================================================*/
ViStatus rssmf_setViReal64Units (ViSession instrSession,
                                 ViString cmd,
                                 ViReal64 value,
                                 ViString units,
                                 ViReal64 min,
                                 ViReal64 max,
                                 ViString parameterName)
{
    ViChar      buffer[BUFFER_SIZE];
    ViStatus    rssmf_status = VI_SUCCESS;

    /* do range checking only if there's difference between max and min */
    if ((max - min) > 1.0e-14)
        if (rssmf_invalidViReal64Range (instrSession, value, min, max) == VI_TRUE)
		{
		    rssmf_instrRange   instrPtr;
		    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
			strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
	        return VI_ERROR_INSTR_PARAMETER;
		}

    sprintf(buffer,"%s %.12f %s",cmd, value, units);

    return rssmf_sendCmd(instrSession, buffer);
}

/*===========================================================================*/
/* Function: Set String                                                      */
/* Purpose:  This function sets command with string parameter.               */
/*===========================================================================*/
ViStatus rssmf_setViString (ViSession instrSession,
                            ViString cmd,
                            ViString value,
                            ViString parameterName)
{
    ViStatus    rssmf_status = VI_SUCCESS;
    ViChar*     wrt_buf;
    ViChar*     p2buf;

    if (value == VI_NULL)
	{
	    rssmf_instrRange   instrPtr;
	    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
        return VI_ERROR_INSTR_PARAMETER;
	}

    wrt_buf = (ViChar*) malloc (strlen(cmd) + strlen(value) + 4 );
    p2buf = wrt_buf;
    p2buf += sprintf(wrt_buf, "%s %s\n", cmd, value);

    GOCHECKERR(viWrite (instrSession, (ViBuf)wrt_buf, (ViUInt32) (p2buf - wrt_buf), VI_NULL));
    GOCHECKERR(rssmf_checkStatus(instrSession));

ERROR_LABEL:

    free(wrt_buf);

    return rssmf_status;
}

/*===========================================================================*/
/* Function: Get String                                                      */
/* Purpose:  This function gets command with string parameter.               */
/*===========================================================================*/
ViStatus rssmf_getViString (ViSession instrSession,
                            ViString cmd,
                            ViString value,
                            ViUInt32 buffer_size,
                            ViString parameterName)
{
    ViStatus    rssmf_status    = VI_SUCCESS;
    ViChar      *pbuffer,
                *p2buffer;
    ViUInt32    count           = 0;
    size_t      length          = 0;

    if (value == NULL)
	{
	    rssmf_instrRange   instrPtr;
	    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
        return VI_ERROR_INSTR_PARAMETER;
	}

    length = strlen (cmd);
    pbuffer = (ViChar*) malloc (length + 1);
    
    memcpy (pbuffer, cmd, length);
    if (pbuffer == NULL)
        return VI_ERROR_ALLOC;
    
    pbuffer[length] = '\n';

    GOCHECKERR (viWrite (instrSession, (ViPBuf)pbuffer, (ViUInt32) (length + 1), VI_NULL));

    free (pbuffer);
    
    pbuffer = NULL;

    GOCHECKERR (rssmf_readDataUnknownLength (instrSession, &pbuffer, &count));
    
    if (pbuffer == NULL)
        return rssmf_status;
    
    GOCHECKERR (rssmf_checkStatus (instrSession));

    if (count > buffer_size)
    {
        rssmf_status = RSSMF_WARNING_SMALL_BUFFER;
        count = buffer_size;
        memcpy (value, pbuffer, (size_t) count);
        value[count] = '\0';
    }
    else
        if ((p2buffer = strpbrk (pbuffer, "\r\n")) != VI_NULL)
        {
            rssmf_status = VI_SUCCESS;
            memcpy (value, pbuffer, (size_t) (p2buffer - pbuffer));
            value[p2buffer - pbuffer] = '\0';
        }

ERROR_LABEL:

    if (pbuffer != NULL)
        free (pbuffer);

    return rssmf_status;
}

/*===========================================================================*/
/* Function: Set Quoted String                                               */
/* Purpose:  This function sets command with quoted string parameter.        */
/*===========================================================================*/
ViStatus rssmf_setViStringQuoted (ViSession instrSession,
                            ViString cmd,
                            ViString quotes,
                            ViString value,
                            ViString parameterName)
{
    ViStatus    rssmf_status = VI_SUCCESS;
    ViChar*     wrt_buf;
    ViChar*     p2buf;

    if (value == VI_NULL)
	{
	    rssmf_instrRange   instrPtr;
	    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
        return VI_ERROR_INSTR_PARAMETER;
	}

    wrt_buf = (ViChar*)malloc(strlen(cmd) + strlen(value) + 2*strlen(quotes) + 4);
    p2buf = wrt_buf;
    p2buf += sprintf(wrt_buf,"%s %s%s%s\n", cmd, quotes, value, quotes);

    GOCHECKERR (viWrite (instrSession, (ViBuf)wrt_buf, (ViUInt32) (p2buf - wrt_buf), VI_NULL));
    GOCHECKERR (rssmf_checkStatus(instrSession));

ERROR_LABEL:

    free (wrt_buf);

    return rssmf_status;
}

/*===========================================================================*/
/* Function: Set String With Channel                                         */
/* Purpose:  This function sets command with string parameter.               */
/*===========================================================================*/
ViStatus rssmf_setViStringPath (ViSession instrSession,
                            ViString cmd,
                            ViInt32 path,
                            ViString value,
                            ViString parameterName)
{
    ViStatus    rssmf_status = VI_SUCCESS;
    ViChar*     wrt_buf;
    ViChar*     p2buf;

    if (value == VI_NULL)
	{
	    rssmf_instrRange   instrPtr;
	    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
        return VI_ERROR_INSTR_PARAMETER;
	}

    wrt_buf = (ViChar*) malloc (strlen(cmd) + strlen(value) + 11);
    p2buf = wrt_buf;
    p2buf += sprintf(wrt_buf, cmd, path);
    p2buf += sprintf(p2buf," %s\n", value);

    rssmf_status = viWrite (instrSession, (ViBuf)wrt_buf, (ViUInt32)(p2buf - wrt_buf), VI_NULL);
    free(wrt_buf);

    CHECKERR(rssmf_status);
    CHECKERR(rssmf_checkStatus(instrSession));

    return rssmf_status;
}

/*===========================================================================*/
/* Function: Get String with Path                                            */
/* Purpose:  This function gets command with string parameter.               */
/*===========================================================================*/
ViStatus rssmf_getViStringPath (ViSession instrSession,
                                ViString cmd,
                                ViInt32 path,
                                ViString value,
                                ViInt32 buffer_size,
                                ViString parameterName)
{
    ViChar      wrt_buf[BUFFER_SIZE];

    sprintf(wrt_buf,cmd, path);

    return rssmf_getViString(instrSession,
                             wrt_buf,
                             value,
                             (ViUInt32)buffer_size,
                             parameterName);
}

/*===========================================================================*/
/* Function: Set Enum                                                        */
/* Purpose:  This function sets command with enumerated parameter.           */
/*===========================================================================*/
ViStatus rssmf_setEnum (ViSession instrSession,
                        ViString cmd,
                        ViInt32 value,
                        ViString list[],
                        ViString parameterName)
{
    ViStatus rssmf_status = VI_SUCCESS;

    CHECKERR (rssmf_checkIndexInList (instrSession, value, list, parameterName));

    return rssmf_setViString (instrSession, cmd, list[value], parameterName);
}

/*===========================================================================*/
/* Function: Get Enum                                                        */
/* Purpose:  This function sets command with enumerated parameter.           */
/*===========================================================================*/
ViStatus rssmf_getEnum (ViSession instrSession,
                        ViString cmd,
                        ViInt32 *value,
                        ViString list[],
                        ViString parameterName)
{
    ViStatus    rssmf_status = VI_SUCCESS;
    ViChar      buffer[BUFFER_SIZE] = "";

    CHECKERR( rssmf_getViString(instrSession, cmd, buffer, BUFFER_SIZE, parameterName));
    CHECKERR( rssmf_getIndexFromString (buffer, list, value));

    return rssmf_status;
}

/*===========================================================================*/
/* Function: Set Enum With Path                                              */
/* Purpose:  This function sets command with enumerated parameter.           */
/*===========================================================================*/
ViStatus rssmf_setEnumPath (ViSession instrSession,
                        ViString cmd,
                        ViInt32 path,
                        ViInt32 value,
                        ViString list[],
                        ViString parameterName)
{
    ViStatus rssmf_status = VI_SUCCESS;

    CHECKERR (rssmf_checkIndexInList (instrSession, value, list, parameterName));

    return rssmf_setViStringPath (instrSession, cmd, path, list[value], parameterName);
}

/*===========================================================================*/
/* Function: Get Enum Path                                                   */
/* Purpose:  This function sets command with enumerated parameter.           */
/*===========================================================================*/
ViStatus rssmf_getEnumPath (ViSession instrSession,
                        ViString cmd,
                        ViInt32 path,
                        ViInt32 *value,
                        ViString list[],
                        ViString parameterName)
{
    ViStatus    rssmf_status = VI_SUCCESS;
    ViChar      buffer[BUFFER_SIZE] = "";

    CHECKERR(rssmf_getViStringPath(instrSession, cmd, path, buffer, (ViInt32)BUFFER_SIZE, parameterName));
    CHECKERR( rssmf_getIndexFromString (buffer, list, value));

    return rssmf_status;

}

/*===========================================================================*/
/* Function: Set ViUInt32                                                    */
/* Purpose:  This function sets command with unsigned integer parameter.     */
/*===========================================================================*/
ViStatus rssmf_setViUInt32 (ViSession instrSession,
                           ViString cmd,
                           ViUInt32 value,
                           ViUInt32 min,
                           ViUInt32 max,
                           ViString parameterName)
{
    ViChar      wrt_buf[BUFFER_SIZE];
    ViStatus    rssmf_status = VI_SUCCESS;

    if ((max - min) > 0)
        if (rssmf_invalidViUInt32Range (instrSession, value, min, max) != 0)
		{
		    rssmf_instrRange   instrPtr;
		    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
			strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
	        return VI_ERROR_INSTR_PARAMETER;
		}

    sprintf (wrt_buf, "%s %u", cmd, value);

    return rssmf_sendCmd (instrSession, wrt_buf);
}

/*===========================================================================*/
/* Function: Get ViUInt32                                                    */
/* Purpose:  This function gets command with unsigned integer parameter.     */
/*===========================================================================*/
ViStatus rssmf_getViUInt32 (ViSession instrSession,
                           ViString cmd,
                           ViUInt32 *value,
                           ViString parameterName)
{
    ViStatus    rssmf_status = VI_SUCCESS;

    if (value == VI_NULL)
	{
	    rssmf_instrRange   instrPtr;
	    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
        return VI_ERROR_INSTR_PARAMETER;
	}

    CHECKERR( viPrintf(instrSession, "%s\n", cmd));
    CHECKERR( viScanf(instrSession, "%u", value));
    CHECKERR( rssmf_checkStatus (instrSession));

    return rssmf_status;
}

/*===========================================================================*/
/* Function: Check Index List                                                */
/* Purpose:  This function checks if the passed index is valid for given list*/
/*===========================================================================*/
ViStatus rssmf_checkIndexInList (ViSession instrSession,
								 ViInt32 index,
                                 ViString list[],
								 ViString parameterName)

{
    ViInt32     i=0;
	ViStatus    rssmf_status = VI_SUCCESS;

    if (index < 0)
	{
	    rssmf_instrRange   instrPtr;
	    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
        return VI_ERROR_INSTR_PARAMETER;
	}

    while (list[i++] != VI_NULL);
    if (index >= i)
	{
	    rssmf_instrRange   instrPtr;
	    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
        return VI_ERROR_INSTR_PARAMETER;
	}

    return VI_SUCCESS;
}

/*===========================================================================*/
/* Function: Get Index From String                                           */
/* Purpose:  This function returns index for string.                         */
/*===========================================================================*/
ViStatus rssmf_getIndexFromString (ViString string,
                                   ViString *list,
                                   ViInt32 *index)
{
    ViStatus    rssmf_status = VI_SUCCESS;
    ViUInt32    i;
    size_t      len;

    for (i = 0; list[i]; i++)
        if ((len = strlen (string)) == strlen (list[i])) {
            if (strncmp(string, list[i], strlen (string)) == 0) break;
        }

    *index = (ViInt32) i;

    return (list[i]) ? rssmf_status : VI_ERROR_INSTR_INTERPRETING_RESPONSE;
}

/*===========================================================================*/
/* Function: Get Catalog                                                     */
/*===========================================================================*/
ViStatus rssmf_getCatalog (ViSession instrSession,
                           ViString cmd,
                           ViUInt32 arraySize,
                           ViChar* array,
                           ViString parameterName)
{
    ViStatus    rssmf_status = VI_SUCCESS;
    ViChar      *pbuffer,
                *p2buffer;
    size_t      count  = 0,
                length = 0;
    ViUInt32    unknown_count = 0;

    if (array == NULL)
	{
	    rssmf_instrRange   instrPtr;
	    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
        return VI_ERROR_INSTR_PARAMETER;
	}

    length = strlen (cmd);
    pbuffer = (ViChar*) malloc (length + 1);
    
    if (pbuffer == NULL)
        return VI_ERROR_ALLOC;
    
    memcpy (pbuffer, cmd, length);
    pbuffer[length] = '\n';

    GOCHECKERR (viWrite (instrSession, (ViPBuf)pbuffer, (ViUInt32) (length + 1), VI_NULL));

    free(pbuffer);
    
    pbuffer = NULL;

    GOCHECKERR (rssmf_readDataUnknownLength (instrSession, &pbuffer, &unknown_count));
    GOCHECKERR (rssmf_checkStatus (instrSession));

    if (count > (size_t) arraySize)
        {
        rssmf_status = RSSMF_WARNING_SMALL_BUFFER;
        count = (size_t)arraySize;
        memcpy (array, pbuffer, count);
        array[count-1] = '\0';
        }
    else
        {
        if ((p2buffer = strchr (pbuffer, ';')) != VI_NULL)
            {
            rssmf_status = VI_SUCCESS;
            memcpy (array, pbuffer, (size_t)(p2buffer - pbuffer));
            array[p2buffer - pbuffer] = '\0';
            }
        else
            {
            memcpy (array, pbuffer, count);
            array[count-1] = '\0';
            }
        }
ERROR_LABEL:

    if (pbuffer != NULL)
        free (pbuffer);

    return rssmf_status;
}

/*===========================================================================*/
/* Function: Get ViReal64 Array                                              */
/*===========================================================================*/
ViStatus rssmf_getViReal64Array (ViSession instrSession,
                            ViString cmd,
                            ViReal64 *value,
                            ViInt32  count,
                            ViString parameterName)
{
    ViStatus    rssmf_status = VI_SUCCESS;
    ViChar*     pbuffer = VI_NULL;
    ViChar*     p2buf;
    ViUInt32    local_count = 0;
    ViInt32     i = 0;

    if (value == VI_NULL)
	{
	    rssmf_instrRange   instrPtr;
	    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
        return VI_ERROR_INSTR_PARAMETER;
	}

    GOCHECKERR (viPrintf(instrSession, "%s\n",cmd));
    GOCHECKERR (rssmf_readDataUnknownLength(instrSession, &pbuffer, &local_count));
    p2buf = pbuffer;

    p2buf = strtok(pbuffer, ",");

    while ((p2buf != NULL) && (i < count))
    {
        value[i++] = atof(p2buf);
        p2buf = strtok(NULL, ",");
    };

    GOCHECKERR (rssmf_checkStatus (instrSession));

ERROR_LABEL:

    free(pbuffer);

    return rssmf_status;
}

/*===========================================================================*/
/* Function: Get ViInt32 Array                                               */
/*===========================================================================*/
ViStatus rssmf_getViInt32Array (ViSession instrSession,
                            ViString cmd,
                            ViInt32 *value,
                            ViInt32  count,
                            ViString parameterName)
{
    ViStatus    rssmf_status = VI_SUCCESS;
    ViChar*     pbuffer = VI_NULL;
    ViChar*     p2buf;
    ViUInt32    local_count = 0;
    ViInt32     i = 0;

    if (value == VI_NULL)
	{
	    rssmf_instrRange   instrPtr;
	    CHECKERR (viGetAttribute (instrSession, VI_ATTR_USER_DATA, &instrPtr));
		strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
        return VI_ERROR_INSTR_PARAMETER;
	}

    GOCHECKERR (viPrintf(instrSession, "%s\n",cmd));
    GOCHECKERR (rssmf_readDataUnknownLength(instrSession, &pbuffer, &local_count));
    p2buf = pbuffer;

    p2buf = strtok(pbuffer, ",");

    while ((p2buf != NULL) && (i < count))
    {
        value[i++] = atol(p2buf);
        p2buf = strtok(NULL, ",");
    };

    GOCHECKERR (rssmf_checkStatus (instrSession));

ERROR_LABEL:

    free(pbuffer);

    return rssmf_status;
}

/*===========================================================================*/
/* Function: Read Data With Unknown length                                   */
/* Purpose:  This function allocates buffer and reads data from the          */
/*           instrument                                                      */
/* WARNING:  You must deallocate buffer, if any pointer has been returned.   */
/*           This function does not work correctly, when you read bin data   */
/*           via serial intf                                                 */
/*===========================================================================*/
ViStatus rssmf_readDataUnknownLength(ViSession instrSession,
                                     /*@null@*/ViChar **outBuffer,
                                     ViUInt32 *count)
{
    ViStatus rssmf_status = VI_SUCCESS;
    ViUInt32 actSize, retCnt = 0;
    ViChar   *buffer = NULL, *p2buffer = NULL;

    *outBuffer = VI_NULL;
    if (count) *count = 0;

    /* Allocate Buffer */
    actSize = BUFFER_SIZE;
    buffer = (ViChar *)calloc((size_t)actSize, (size_t) sizeof(ViChar));
    if (buffer == VI_NULL) return VI_ERROR_ALLOC;

    p2buffer = buffer;
    do
    {
        if ((rssmf_status = viRead (instrSession, (ViPBuf)p2buffer, BUFFER_SIZE, &retCnt)) < 0)
        {
            free(buffer);
            *outBuffer = VI_NULL;
            return rssmf_status;
        }
        if (retCnt == BUFFER_SIZE) /* Buffer was small, reallocate it */
        {
            actSize += BUFFER_SIZE;
            buffer = (ViChar *)realloc(buffer, (size_t) actSize);
            if (buffer == VI_NULL) return VI_ERROR_ALLOC;
            p2buffer = buffer + (actSize - BUFFER_SIZE); /* Set pointer to end of data in rallocated buffer */
        }
        else
        {
            *(p2buffer+retCnt) = '\0'; /* For sure set end of string at the end of data */
        }
    } while (rssmf_status == VI_SUCCESS_MAX_CNT); /* if buffer was small, continue with next read */

    *outBuffer = buffer;
    if (count != NULL) *count = (actSize - BUFFER_SIZE) + retCnt;

    return rssmf_status;
}

/*===========================================================================*/
/* Function: Read Data To File                                               */
/* Purpose:  This function read data from the instrument and                 */
/*           write it to a user specified file.                              */
/*===========================================================================*/
ViStatus rssmf_readDataToFile(
    ViSession   instrSession,
    ViString    cmd,
    ViString    destination
)
{
    ViStatus    rssmf_status    = VI_SUCCESS;
    ViUInt16    intfType        = VI_INTF_GPIB;
	ViUInt64    length          = 0;
	ViInt32     i = 0;
    ViChar      buffer[BUFFER_SIZE],
                *pbuffer = NULL;

    CHECKERR (viGetAttribute (instrSession, VI_ATTR_INTF_TYPE, &intfType));

    if (intfType == VI_INTF_GPIB)
        CHECKERR (viWrite (instrSession, (ViBuf) ":SYST:COMM:GPIB:LTER EOI\n", 25, VI_NULL));

    sprintf (buffer, "%s\n", cmd);

    CHECKERR (viWrite (instrSession, (ViBuf) buffer, (ViUInt32) strlen(buffer), VI_NULL));
    CHECKERR (viRead (instrSession, (ViBuf) buffer, 1, NULL)); 
    
	if (buffer[0] == '#')
	{   /* Binary data */
		/* Disable termination character */
		CHECKERR(viSetAttribute(instrSession, VI_ATTR_TERMCHAR_EN, VI_FALSE));

		/* Parse header */
		CHECKERR(viRead(instrSession, (ViBuf)buffer, 1, NULL));

		buffer[1] = '\0';
		if (buffer[0] == '(')
		{
			for (i = 0; i < 20; i++)
			{
				CHECKERR(viRead(instrSession, (ViBuf)buffer + i, 1, NULL));

				if (buffer[i] == ')')
				{
					buffer[i] = '\0';
					sscanf(buffer, "%lld", &length);
					break;
				}
			}
			if (i == 20)
			{
				//couldn't find the ')' character
				CHECKERR(VI_ERROR_INSTR_INTERPRETING_RESPONSE);
			}
		}
		else
		{
			sscanf(buffer, "%lld", &length);
			if (length > 9 || length < 1)
			{
				//expected number 1 to 9
				CHECKERR(VI_ERROR_INSTR_INTERPRETING_RESPONSE);
			}

			CHECKERR(viRead(instrSession, (ViBuf)buffer, (ViUInt32) length, NULL));
			buffer[length] = '\0';
			sscanf(buffer, "%lld", &length);
		}

		CHECKERR(rssmf_ReadToFile(instrSession, destination, length, RS_VAL_TRUNCATE));
		/* Flush garbage */
		viScanf(instrSession, "%*[^\r\n]");
		if (rssmf_status > 0) rssmf_status = VI_SUCCESS;
    }
    else
    {
        CHECKERR(VI_ERROR_INSTR_INTERPRETING_RESPONSE);         
    }
	
    viSetAttribute (instrSession, VI_ATTR_TERMCHAR_EN, VI_TRUE);

    if (intfType == VI_INTF_GPIB)
        if ((rssmf_status = viWrite (instrSession, (ViBuf) ":SYST:COMM:GPIB:LTER STAN\n", 26, VI_NULL)) < VI_SUCCESS)
            {
            free (pbuffer);
            return rssmf_status;
            }

    if (pbuffer) /* If not NULL pointer, then deallocate memory. */
        free (pbuffer);

    rssmf_status = rssmf_checkStatus (instrSession);
    return rssmf_status;
}

ViStatus rssmf_ParameterError (ViSession io, ViString parameterName)
{
	ViStatus    rssmf_status = VI_SUCCESS;
	rssmf_instrRange   instrPtr;
	CHECKERR (viGetAttribute (io, VI_ATTR_USER_DATA, &instrPtr));
	strncpy (instrPtr->parameterErrorMessage, parameterName, 255);
	return VI_ERROR_INSTR_PARAMETER;
}

ViStatus rssmf_ReadToFile (ViSession instrSession,
                        ViConstString filename,
                        ViUInt64 maxBytesToRead,
                        ViInt32 fileAction)
{
    ViStatus    error               = VI_SUCCESS;
    ViUInt64    readBytes           = maxBytesToRead,
                bytesWrittenFile    = 0;
    FILE        *targetFile         = NULL;
    ViByte*     buffer = NULL;
	ViUInt32 	bytesReadInstr = 0;

    if ((targetFile = fopen (filename, (fileAction == RS_VAL_APPEND) ? "ab" : "wb")) == NULL)
        return VI_ERROR_INSTR_FILE_OPEN;

    buffer = (ViByte*) malloc(RS_IO_BUFFER_SIZE);
	CHECKNULL (buffer);

    memset ((void *) buffer, 0, (size_t) RS_IO_BUFFER_SIZE);

    for (;;)
        {
        if (readBytes > RS_IO_BUFFER_SIZE)
            error = viRead (instrSession, buffer, RS_IO_BUFFER_SIZE, &bytesReadInstr);
        else
            error = viRead (instrSession, buffer, (ViUInt32) readBytes, &bytesReadInstr);

        bytesWrittenFile = (ViUInt32) fwrite (buffer, sizeof (ViByte), (size_t)bytesReadInstr, targetFile);

        if (bytesWrittenFile < bytesReadInstr)
            error = VI_ERROR_INSTR_FILE_WRITE;

        if ((readBytes <= RS_IO_BUFFER_SIZE) || (error <= 0) || (error == VI_SUCCESS_TERM_CHAR))
            break;

        readBytes -= RS_IO_BUFFER_SIZE;
        }

    (void) fclose (targetFile);

  if (buffer) free (buffer);

    return error;
}

ViStatus rssmf_WriteFromFile (ViSession instrSession,
                           ViConstString filename,
                           ViUInt64 maxBytesToWrite,
                           ViUInt64 byteOffset)
{
    ViStatus    error           = VI_SUCCESS;
    ViUInt64      bytesRead       = 0,
                writeBytes      = maxBytesToWrite;
    FILE        *sourceFile;
    ViByte*     buffer = NULL;

    if ((sourceFile = fopen (filename, "rb")) == NULL)
        return VI_ERROR_INSTR_FILE_OPEN;

    if (fseek (sourceFile, (ViUInt32) byteOffset, SEEK_SET) != VI_SUCCESS)
        return VI_ERROR_INSTR_FILE_READ;

    (void) viSetAttribute (instrSession, VI_ATTR_SEND_END_EN, VI_FALSE);
    
	buffer = (ViByte*) malloc(RS_IO_BUFFER_SIZE);
	CHECKNULL (buffer);
    memset ((void *) buffer, 0, (size_t) RS_IO_BUFFER_SIZE);

    while (feof (sourceFile) == 0)
    {
        bytesRead = (ViUInt32) fread (buffer, sizeof (ViByte), (size_t) RS_IO_BUFFER_SIZE, sourceFile);
        if ((writeBytes > RS_IO_BUFFER_SIZE) && (bytesRead == RS_IO_BUFFER_SIZE))
        {
            error = viWrite (instrSession, buffer, RS_IO_BUFFER_SIZE, NULL);
            writeBytes -= RS_IO_BUFFER_SIZE;
            if (error < 0)
                break;
        }
        else
        {
            (void) viSetAttribute (instrSession, VI_ATTR_SEND_END_EN, VI_TRUE);
			error = viWrite (instrSession, buffer, (ViUInt32) ((bytesRead < writeBytes) ? bytesRead : writeBytes), NULL);
            break;
        }
    }

	(void) viSetAttribute (instrSession, VI_ATTR_SEND_END_EN, VI_TRUE);
	(void) fclose (sourceFile);
	if (buffer) free (buffer);

    return error;
}
